use v5.10;
use utf8;

package Local::DocBook;
use strict;
use base 'Pod::PseudoPod';

use warnings;
no warnings;

use subs qw();
use vars qw($VERSION);

use Carp;

$VERSION = '0.10';

sub DEBUG () { 0 }

=head1 NAME

Pod::PseudoPod::DocBook - Turn Pod into DocBook

=head1 SYNOPSIS

	use Pod::PseudoPod::DocBook;

=head1 DESCRIPTION

***THIS IS ALPHA SOFTWARE. MAJOR PARTS WILL CHANGE***

I wrote just enough of this module to get my job done, and I skipped every
part of the specification I didn't need while still making it flexible enough
to handle stuff later.

=head2 The style information

I don't handle all of the complexities of styles, defining styles, and
all that other stuff. There are methods to return style names, and you
can override those in a subclass.

=cut

=over 4

=item document_header

This is the start of the document that defines all of the styles. You'll need
to override this. You can take this directly from

=cut

sub add_xml_tag {
	my( $self, $stuff ) = @_;

	$self->add_to_pad( $stuff );
	$self->emit;
	}

sub add_data {
	my( $self, $stuff ) = @_;

	$self->add_to_pad( $stuff );
	$self->escape_and_emit;
	}

sub escape_and_emit {
	my( $self ) = @_;

	my $pad = $self->get_pad;

	$self->{$pad} =~ s/\s+\z//;

	$self->{$pad} =~ s/\&
		(?!
			(?:amp|lt|gt|\x23[a-f0-9]+);
		)/&amp;/xgi;

	$self->{$pad} =~ s/</&lt;/g;
	$self->{$pad} =~ s/>/&gt;/g;

	$self->emit;
	}

sub add_to_pad {
	my( $self, $stuff ) = @_;
	my $pad = $self->get_pad;
	$self->{$pad} .= $stuff;
	}

sub clear_pad {
	my( $self ) = @_;

	my $pad = $self->get_pad;
	$self->{$pad} = '';
	}

sub set_title           { $_[0]->{title} = $_[1] }
sub set_chapter         { $_[0]->{chapter} = $_[1] }

sub title               { $_[0]->{title}      }
sub chapter             { $_[0]->{chapter}    }

sub section             { $_[0]->{section}    }
sub subsection          { $_[0]->{subsection} }
sub subsubsection       { $_[0]->{subsubsection} }
sub subsubsubsection    { $_[0]->{subsubsubsection} }
sub subsubsubsubsection { $_[0]->{subsubsubsubsection} }

sub document_header {
	my $string = <<"XML";
<?xml version="1.0" encoding="UTF-8"?>
<!DOCTYPE chapter PUBLIC "-//OASIS//DTD DocBook XML V4.5//EN"
	"http://www.oasis-open.org/docbook/xml/4.5/docbookx.dtd">
<!-- created by brian's private converter -->
XML

	my $id = join '-', $_[0]->title, $_[0]->chapter;

	my $filename = $_[0]->{source_filename};
	$_[0]->{root_tag} = do {
		    if( $filename =~ /app/ ) { 'appendix' }
		elsif( $filename =~ /ch00/ ) { 'preface'  }
		else                         { 'chapter'  }
		};

	$string .= <<"XML";
<$_[0]->{root_tag} id="$id">

XML
	}

=item document_footer

=cut

sub document_footer {
	my $string = '';

	foreach my $section (  @{ $_[0]->{in_section} } )
		{
		$string .= "</sect$section>\n";
		}

	$string .= <<"XML";
</$_[0]->{root_tag}>
XML
	}

=back

=head2 The Pod::Simple mechanics

Everything else is the same stuff from C<Pod::Simple>.

=cut

use Data::Dumper;

sub new {
	my $self = $_[0]->SUPER::new();
	$self->{accept_targets}{table}++;
	$self->{accept_targets}{figure}++;
	$self->{accept_targets}{production}++;
	$self->accept_codes( 'Y' );
	$self->accept_codes( 'V' );
	$self->accept_codes( 'P' );
	$self;
	}

sub emit {
	my $pad = $_[0]->get_pad;
	print {$_[0]->{'output_fh'}} $_[0]->{$pad};
	$_[0]->clear_pad;
	return;
	}

sub get_pad {
	# flow elements first
	   if( $_[0]{in_A}      )     { 'ref_text'    }
	elsif( $_[0]{in_L}      )     { 'link_text'   }
	elsif( $_[0]{in_M}      )     { 'module_text' }
	elsif( $_[0]{in_P}      )     { 'entity_text' }
	elsif( $_[0]{in_U}      )     { 'url_text'    }
	elsif( $_[0]{in_V}      )     { 'var_text'    }
	elsif( $_[0]{in_X}      )     { 'index_text'  }
	elsif( $_[0]{in_figure} )     { 'figure_text' }
	elsif( $_[0]{in_production} ) { 'remark_text' }
	# then block elements
	# finally the default
	else                          { 'scratch'     }
	}

sub get_text_from_current_pad {
	my $pad = $_[0]->get_pad;
	my $text = $_[0]->{$pad};
	}

sub start_Document {
	$_[0]->{in_section} = [];
	$_[0]->add_to_pad( $_[0]->document_header );
	$_[0]->emit;
	}

sub end_Document {
	$_[0]->add_to_pad( $_[0]->document_footer );
	$_[0]->emit;
	}

sub _header_start {
	my( $self, $level ) = @_;

	$self->{in_header} = 1;
	if( $level )
		{
		LEVEL: {
			if( eval { @{ $self->{in_section} } } and $self->{in_section}[0] >= $level )
				{
				my $tag = shift @{ $self->{in_section} };
				$self->add_xml_tag( "</sect$tag>\n" );
				redo LEVEL;
				}
			last LEVEL;
			}

		my @parts = qw(title chapter section);
		push @parts, 'subsection'    if $level > 1;
		push @parts, 'subsubsection' if $level > 2;
		push @parts, 'subsubsubsection' if $level > 3;

		@parts = map { $self->$_() } @parts;

		my $id = join '-', @parts;
		my $tag = qq|\n<sect$level id="$id">\n|;

		$self->add_xml_tag( $tag );
		unshift @{ $self->{in_section} }, $level;
		}

	$self->add_xml_tag( qq|<title>| );
	}

sub _header_end {
	my( $self, $level ) = @_;
	$self->{in_header} = 0;

	$self->add_xml_tag( "</title>\n" );
	}

sub start_head0     { $_[0]->_header_start( 0 ); }
sub end_head0       { $_[0]->_header_end( 0 );   }

sub end_head1       { $_[0]->_header_end( 1 );   }
sub start_head1     {
	$_[0]->{section}++;
	$_[0]->{subsection} = 0;

	$_[0]->_header_start( 1 );
	}

sub end_head2       { $_[0]->_header_end( 2 );   }
sub start_head2     {
	$_[0]->{subsection}++;
	$_[0]->{subsubsection} = 0;

	$_[0]->_header_start( 2 );
	}

sub start_head3     { 
	$_[0]->{subsubsubsection} = 0;
	$_[0]->{subsubsection}++;
	$_[0]->_header_start( 3 ); 
	}
sub end_head3       { $_[0]->_header_end( 3 );   }

sub start_head4     { 
	$_[0]->{subsubsubsubsection} = 0;
	$_[0]->{subsubsubsection}++;
	$_[0]->_header_start( 4 ); 
	}
sub end_head4       { $_[0]->_header_end( 4 );   }

sub end_non_code_text {
	my $self = shift;

	$self->make_curly_quotes;

	$self->emit;
	}

sub start_Para {
	my $self = shift;

	return if( $self->{in_figure} or $self->{in_production} );

	if( $self->{in_item_text} ) {
		$self->add_xml_tag( qq|<term>| );				
		}
	else {
		$self->add_xml_tag( qq|<para>| );
		}

	$self->escape_and_emit;

	$self->{'in_para'} = 1;
	}

sub end_Para {
	my $self = shift;

	return if( $self->{in_figure} or $self->{in_production} );

	if( $self->{in_item_text} ) {
		$self->add_xml_tag( qq|</term><listitem>| );				
		$self->{'in_item_text'} = 0;
		}
	else {
		$self->add_xml_tag( qq|</para>\n\n| );
		$self->{'in_para'} = 0;
		}

	$self->end_non_code_text;

	}

sub start_Verbatim {
	$_[0]{'in_verbatim'} = 1;
	my $sequence = ++$_[0]{'verbatim_sequence'};
	my $chapter  = $_[0]->chapter;

	$_[0]->add_xml_tag( qq|\n<programlisting format="linespecific" id="I_programlisting_${chapter}_tt${sequence}" xml:space="preserve">| );
	$_[0]->emit;
	}

sub end_Verbatim {
	my $self = shift;

	# get rid of all but one trailing newline
	#$self->escape_and_emit;

	$self->add_xml_tag( "</programlisting>\n\n" );

	$self->{'in_verbatim'} = 0;
	}

sub not_implemented { croak "Not implemented! " . (caller(1))[3] }

=head2 List handling

=cut

sub in_item_list { scalar @{ $_[0]->{list_levels} } }
sub add_list_level_item {
	${ $_[0]->{list_levels} }[-1]{item_count}++;
	}
sub is_first_list_level_item {
	${ $_[0]->{list_levels} }[-1]{item_count} == 0;
	}

sub start_list_level {
	my $self = shift;

	push @{ $self->{list_levels} }, { item_count => 0 };
	}

sub end_list_level {
	my $self = shift;

	pop @{ $self->{list_levels} };
	}

sub start_item {
	my( $self, $tag ) = @_;
	$tag //= 'listitem'; #/

	
	unless( $self->is_first_list_level_item ) {
		$self->add_xml_tag( "</listitem>\n" );		
		$self->add_xml_tag( "</varlistentry>\n\n" )
			if $tag eq 'varlistentry';
		}

	$self->add_list_level_item;
	$self->add_xml_tag( "<$tag>\n" );

	$self->start_Para;
	}

sub end_item {
	my $self = shift;
	$self->end_Para;
	$self->{in_item} = 0;
	}

sub start_item_bullet { $_[0]->start_item }
sub start_item_number { $_[0]->start_item }
sub start_item_block  { $_[0]->start_item }
sub start_item_text   { $_[0]->{in_item_text}++; $_[0]->start_item( 'varlistentry' ) }

sub end_item_bullet { $_[0]->end_item }
sub end_item_number { $_[0]->end_item }
sub end_item_block  { $_[0]->end_item }
sub end_item_text   { $_[0]->end_item }

sub start_over {
	my( $self, $tag ) = @_;
	$tag //= '<itemizedlist>'; #/

	$self->add_xml_tag( $tag );

	$self->start_list_level;
	}

sub end_over {
	my( $self, $tag ) = @_;
	$tag //= '</itemizedlist>'; #'

	$self->{last_thingy}  = 'item_list';
	$self->end_non_code_text;
	$self->add_xml_tag( "</listitem>\n" );
	$self->add_xml_tag( "</varlistentry>\n\n" )
		if $tag eq '</variablelist>';

	$self->add_to_pad( "$tag\n\n" );
	$self->end_list_level;
	$self->emit;
	}

sub start_over_bullet { $_[0]->start_over( '<itemizedlist>' ) }
sub start_over_text   { $_[0]->start_over( '<variablelist>' ) }
sub start_over_block  { $_[0]->start_over( '<itemizedlist>' ) }
sub start_over_number { $_[0]->start_over( '<orderedlist>'  ) }

sub end_over_bullet { $_[0]->end_over( '</itemizedlist>' ) }
sub end_over_text   { $_[0]->end_over( '</variablelist>' ) }
sub end_over_block  { $_[0]->end_over( '</itemizedlist>' ) }
sub end_over_number { $_[0]->end_over( '</orderedlist>'  ) }

sub _ponder_over {
  my ($self,$para,$curr_open,$paras) = @_;
  return 1 unless @$paras;
  my $list_type;

  if($paras->[0][0] eq '=item') { # most common case
    $list_type = $self->_get_initial_item_type($paras->[0]);

  } elsif($paras->[0][0] eq '=back') {
    # Ignore empty lists by default
    if ($self->{'parse_empty_lists'}) {
      $list_type = 'empty';
    } else {
      shift @$paras;
      return 1;
    }
  } elsif($paras->[0][0] eq '~end') {
    $self->whine(
      $para->[1]{'start_line'},
      "=over is the last thing in the document?!"
    );
    return 1; # But feh, ignore it.
  } else {
    $list_type = 'block';
  }
  $para->[1]{'~type'} = $list_type;
  push @$curr_open, $para;
   # yes, we reuse the paragraph as a stack item
  
  my $content = join ' ', splice @$para, 2;
  my $overness;
  if($content =~ m/^\s*$/s) {
    $para->[1]{'indent'} = 4;
  } elsif($content =~ m/^\s*((?:\d*\.)?\d+)\s*$/s) {
    no integer;
    $para->[1]{'indent'} = $1;
    if($1 == 0) {
      $self->whine(
        $para->[1]{'start_line'},
        "Can't have a 0 in =over $content"
      );
      $para->[1]{'indent'} = 4;
    }
  } else {
    $self->whine(
      $para->[1]{'start_line'},
      "=over should be: '=over' or '=over positive_number'"
    );
    $para->[1]{'indent'} = 4;
  }
  DEBUG > 1 and print "=over found of type $list_type\n";
  
  $self->{'content_seen'} ||= 1;
  $self->_handle_element_start((my $scratch = 'over-' . $list_type), $para->[1]);

  return;
}

sub _get_item_type {       # mutates the item!!
  my($self, $para) = @_;
  return $para->[1]{'~type'} if $para->[1]{'~type'};

  # Otherwise we haven't yet been to this node.  Maybe alter it...
  
  my $content = join "\n", @{$para}[2 .. $#$para];

  if($content =~ m/^\s*\*\s*$/s or $content =~ m/^\s*$/s) {
    # Like: "=item *", "=item   *   ", "=item"
    splice @$para, 2; # so it ends up just being ['=item', { attrhash } ]
    $para->[1]{'~orig_content'} = $content;
    return $para->[1]{'~type'} = 'bullet';

  } elsif($content =~ m/^\s*\*\s+(.+)/s) {  # tolerance
  
    # Like: "=item * Foo bar baz";
    $para->[1]{'~orig_content'}      = $content;
    $para->[1]{'~_freaky_para_hack'} = $1;
    DEBUG > 2 and print " Tolerating $$para[2] as =item *\\n\\n$1\n";
    splice @$para, 2; # so it ends up just being ['=item', { attrhash } ]
    return $para->[1]{'~type'} = 'bullet';

  } elsif($content =~ m/^\s*(\d+)\.?\s*$/s) {
    # Like: "=item 1.", "=item    123412"
    
    $para->[1]{'~orig_content'} = $content;
    $para->[1]{'number'} = $1;  # Yes, stores the number there!

    splice @$para, 2; # so it ends up just being ['=item', { attrhash } ]
    return $para->[1]{'~type'} = 'number';
    
  } else {
    # It's anything else.
    $para->[1]{'~orig_content'} = $content;
    return $para->[1]{'~type'} = 'text';

  }
}

=head2 Figure handling

=cut

sub start_figure 	{
	my( $self, $flags ) = @_;
	$self->{in_figure} = 1;
	$self->{figure_title} = $flags->{title};
	}

=pod

   <figure id="FIG3-1_ID_HERE">
     <title>FIG3-1_TITLE_HERE</title>
     <mediaobject>
       <imageobject role="web">
         <imagedata fileref="figs/web/lnp5_0301.png" format="PNG"/>
       </imageobject>
     </mediaobject>
   </figure>

=cut

sub get_next_figure_id {
	my( $self ) = @_;
	$self->title . '-' . $self->chapter .
		'-FIGURE-' . ++$_[0]{'figure_count'};

	}

sub get_filenames {
	my( $self ) = @_;
	my $filenames = $self->get_text_from_current_pad;
	$self->clear_pad;
	
	$filenames =~ s/\A\s+//;
	$filenames =~ s/\s+\z//;
	
	my @filenames = split /\s+/, $filenames;
	}
	
sub end_figure {
	my( $self, $flags ) = @_;

	state $inline_images = { map { $_, 1 } qw(
		figs/incoming/dollarat.gif
		figs/incoming/dollarbang.gif
		figs/incoming/dollarquestion.gif
		figs/incoming/dollarunderscore.gif
		figs/incoming/taint.gif
		figs/incoming/taintgrey.gif
		figs/incoming/xarg.gif
		figs/incoming/xro.gif
		figs/incoming/xt.gif
		figs/incoming/xu.gif
		figs/incoming/xwide.gif	
		) };

	my @filenames = $self->get_filenames;

	foreach my $file ( @filenames ) {	
		my $method = do {
			if( exists $inline_images->{$file} ) { 'make_inline_media_object' }
			else                                 { 'make_figure' }
			};
		
		$self->$method( $file );
		}

	$self->{figure_title} = '';
	$self->{in_figure} = 0;
	}

sub get_image_type {
	my( $self, $file ) = @_;
	my( $format ) = map { uc } ($file =~ /(gif|p(?:ng|df))\z/ig);
	$format;
	}

sub get_image_width {
	use Image::Info qw(image_info dim);
	my( $self, $file ) = @_;
	my $info = image_info( $file );
	my( $width ) = dim( $info );
	$width;	
	}

sub make_inline_media_object {
	my( $self, $file ) = @_;

	my $file_type = $self->get_image_type( $file );
	my $width     = $self->get_image_width( $file );
	my $role      = 'right';
	my $align     = 'left';
	my $role      = 'web';

	$self->add_xml_tag( <<"XML" );
     <inlinemediaobject>
         <imageobject role="$role">
           <imagedata align="$align"
                      fileref="../$file"
                      format="$file_type" width="$width" />
         </imageobject></inlinemediaobject>
XML
	
	}
	
sub make_informal_figure {
	my( $self, $file ) = @_;

	my $id = $self->get_next_figure_id;
	my $file_type = $self->get_image_type( $file );
	my $width     = $self->get_image_width( $file );
	my $role      = 'web';

	$self->add_xml_tag( <<"XML" );

    <informalfigure id="$id">
       <mediaobject>
         <imageobject role="$role">
           <imagedata fileref="../$file" format="$file_type" width="$width"/>
         </imageobject>
       </mediaobject>
    </informalfigure>

XML
	}

sub make_figure {
	my( $self, $file ) = @_;

	my $id = $self->get_next_figure_id;
	my $file_type = $self->get_image_type( $file );
	my $width     = $self->get_image_width( $file );
	my $role      = 'web';

	$self->add_xml_tag( <<"XML" );

   <figure id="$id">
     <title>$self->{figure_title}</title>
     <mediaobject>
       <imageobject role="$role">
         <imagedata fileref="../$file" format="$file_type" width="$width"/>
       </imageobject>
     </mediaobject>
   </figure>

XML
	}

sub start_table {
	my( $self, $flags ) = @_;
	$_[0]{'in_bodyrow'} = 0;
	$_[0]{'in_headrow'} = 0;
	$_[0]->{rows} = 0;

	my $id = $self->title . '-' . $self->chapter .
		'-TABLE-' . ++$_[0]{'table_count'};
	$self->add_xml_tag(
		qq|<table id="$id">\n| .
		qq|<title>$flags->{'title'}</title>\n| .
		qq|<tgroup cols="2">\n|
		);
	}

sub end_table      {
	$_[0]{'in_bodyrow'} = 0;
	$_[0]{'in_headrow'} = 0;
	$_[0]->{rows} = 0;
	$_[0]->add_xml_tag( "</tbody></tgroup></table>\n\n" );
	}

sub start_headrow  {
	$_[0]->{rows} = 0;
	$_[0]{entry} = 0;
	$_[0]{'in_headrow'} = 1;
	$_[0]{'in_bodyrow'} = 0;
	$_[0]->add_xml_tag( '<thead>' );
	}

sub start_bodyrows {
	$_[0]->end_row;
	$_[0]->add_xml_tag( '</thead>' );
	$_[0]->{rows} = 0;
	$_[0]{entry} = 0;
	$_[0]->add_xml_tag( '<tbody>' );
	$_[0]{'in_bodyrow'} = 1;
	$_[0]{'in_headrow'} = 0;
	}

sub start_row {
	$_[0]->{rows}++;
	$_[0]->{entry} = 0;

	$_[0]->add_xml_tag( qq(<row>\n) );
	}
sub end_row {
	$_[0]->add_xml_tag( qq(</entry>\n</row>\n) ) if $_[0]{entry};
	}

sub start_production {
	my( $self ) = @_;
	$self->emit;
	$_[0]{'in_production'} = 1;
	}

sub end_production      {
	my $text = $_[0]->get_text_from_current_pad;
	$_[0]->clear_pad;

	$_[0]->add_xml_tag( '<remark>' );
	$_[0]->add_to_pad( $text );
	$_[0]->add_xml_tag( '</remark>' );
	$_[0]->emit;
	$_[0]{'in_production'} = 0;
	}

sub start_cell {
	$_[0]->start_row if $_[0]{rows} == 0;
	$_[0]->{entry}++;
	$_[0]->add_xml_tag( qq(</entry>\n) ) if $_[0]->{entry} > 1;
	$_[0]->add_xml_tag( qq(\t<entry align="left" valign="top">) ) 
	}
sub end_cell   { }

sub start_A { $_[0]->emit; $_[0]->{in_A} = 1 }
sub end_A   {
	my $old_text = $_[0]->get_text_from_current_pad;
	$_[0]->clear_pad;

	my $text = do {
		local $_ = $old_text;
		if   ( /\A\d+\z/          ) { sprintf '%02d', $old_text }
		elsif( /\A[abc]\z/i       ) { "app" . lc($old_text)     }
		elsif( /\aa\.(\d+)/i      ) { "appa-$1" }
		elsif( /\Achp?-(\d+)/i    ) { sprintf '%02d', $1; }
		elsif( /\APERL4-CH(\d+)/i ) { sprintf '%02d', $1; }
		elsif( /\AP(\d+)/i        ) { sprintf 'p%d', $1   }            # PART
		elsif( /\AG(?:lossary)?\z/i ) { 'glossary' }                   # Glossary
		elsif( /\AF(\d+)\.(\d+)/i ) { sprintf '%02d-FIGURE-%d', $1, $2 } # <figure id="camel-31-FIGURE-1">
		elsif( /\AT(\d+)\.(\d+)/i ) { sprintf '%02d-TABLE-%d', $1, $2 }  # <figure id="camel-31-TABLE-1">
		elsif( /\AI(.*)/i )         { sprintf 'index-%s', $1 }  # reference to index
		else {
			$_[0]->my_warn( "Found odd reference [$old_text]" );
			();
			}
		};

	my $link = join '-', $_[0]->title, $text;
	if( $ENV{STUB_REFS} ) {
		$_[0]->my_warn(  "Skipping link to ref $link because STUB_REFS" );
		$_[0]->add_xml_tag( '!!REF!!' );
		}
	elsif( not defined $text ) {
		$_[0]->my_warn( "Skipping link to ref $text because I don't recognize it" );
		$_[0]->add_xml_tag( '!!REF!!' );
		}
	else {
		$_[0]->add_xml_tag( qq|<xref linkend="$link" />| );
		}

	$_[0]->emit;
	$_[0]->{in_A} = 0;
	}

sub end_B   { $_[0]->add_xml_tag( '</command>' ); $_[0]->{in_B} = 0; }
sub start_B {
	$_[0]->add_xml_tag( '<command>' );
	$_[0]->{in_B} = 1;
	}

sub end_C   { $_[0]->add_xml_tag( '</literal>' ); $_[0]->{in_C} = 0; }
sub start_C { $_[0]->add_xml_tag( '<literal moreinfo="none">' ); $_[0]->{in_C} = 1; }

sub end_F   { $_[0]->add_xml_tag( '</filename>' ) }
sub start_F { $_[0]->add_xml_tag( '<filename>' )  }

sub start_G { $_[0]->emit; $_[0]->{in_G} = 1 }
sub end_G   {
	my $pad = $_[0]->get_pad;
	my $original = $_[0]->{$pad};
	
	my( $text, $link_text ) = split /\|/, $original;
	$link_text //= $text; # /
	
	$_[0]->clear_pad;

	my $link = $_[0]->make_ref( $link_text );

	$_[0]->add_xml_tag( qq|<glossterm linkend="$link">$text</glossterm>| );
	$_[0]->add_xml_tag( '<indexterm>' );

	$_[0]->add_xml_tag( '<primary>' );
	$_[0]->add_to_pad( $text );
	$_[0]->escape_and_emit;
	$_[0]->add_xml_tag( '</primary>' );
	$_[0]->add_xml_tag( '</indexterm>' );
	
	$_[0]->emit;
	$_[0]->{in_G} = 0;
	}

sub link_already_defined {
	my( $self, $link ) = @_;

	exists $self->{links}{$link};	
	}

sub make_ref {
	my( $self, $text ) = @_;
	
	$text = $text;
	$text =~ s/\s+/-/gr;
	$text =~ s/[^a-z]/-/igr;
	}

sub end_I   { $_[0]->add_xml_tag( '</emphasis>' ) }
sub start_I { $_[0]->add_xml_tag( '<emphasis>' )  }

=pod

<footnote id="intermediateperl-CHP-3-FN2">
          <para>Although we don’t go into here, the <literal moreinfo="none">Module::CoreList</literal> module has the lists of
          which modules came with which versions of Perl, along with other
          historical data.</para>
        </footnote>

<ulink role="orm:hideurl"
url="http://www.guardian.co.uk/film/2006/
sep/22/londonfilmfestival2006.londonfilmfestival">Automavision</ulink>
=cut

sub _treat_Ls { }
sub end_L   {
	my $original = $_[0]->get_text_from_current_pad;
	$_[0]->clear_pad;
	my $emphasis;

	my( $link, $text ) = do {
		if( $original =~ /(\w+)\(\d\)/p ) { #manpage
			( ${^MATCH}, "some link" );
			}
		elsif( $original =~ /\A(perl.*?)\|(#.+)/ ) { # anchor into page
			my $page = $1;
			my $anchor = $2 =~ s/-/ /gr;
			$anchor =~ s/#//;
			$anchor =~ s/%22//g;
			$emphasis = 1;
			( "http://perldoc.perl.org/$page.html$2", "“$anchor” in $page" );
			}
		elsif( $original =~ /\A(perl.*)/ ) { # anchor into page
			$emphasis = 1;
			( "http://perldoc.perl.org/$1.html", $1 );
			}
		elsif( $original =~ /(.*?)\|(https?:.*?)/i ) { # page with user supplied link text
			( $2, $1 );
			}
		else {
			$_[0]->my_warn( "No URL for [$original]" );
			$original
			}
		};

	$_[0]->add_xml_tag( '<emphasis>' ) if $emphasis;
	$_[0]->add_xml_tag( qq|<ulink role="orm:hideurl" url="$link">| ) if $link;
	$_[0]->add_to_pad( $text );
	$_[0]->add_xml_tag( '</ulink>' ) if $link;
	$_[0]->add_xml_tag( '</emphasis>' ) if $emphasis;
	$_[0]->emit;
	$_[0]->{in_L} = 0;
	}
sub start_L {
	$_[0]->emit;
	$_[0]->{in_L} = 1;
	}

sub start_M {
	$_[0]->emit;
	$_[0]->{in_M} = 1;
	$_[0]->clear_pad;
	}
sub end_M   {
	my( $self ) = @_;
	my $module = $self->get_text_from_current_pad;
	$_[0]{'in_M'} = 0;
	
	$self->add_xml_tag( '<literal moreinfo="none">' );
	$self->add_to_pad( $module );
	$self->add_xml_tag( '</literal>' );


	$_[0]->add_xml_tag( '<indexterm type="module">' );

	$_[0]->add_xml_tag( '<primary>' );
	$_[0]->add_to_pad( $module );
	$_[0]->escape_and_emit;
	$_[0]->add_xml_tag( '</primary>' );

	$_[0]->add_xml_tag( '</indexterm>' );

	}

sub end_N   { $_[0]->add_xml_tag( '</para></footnote>' ); $_[0]->{in_N} = 0; }
sub start_N {
	$_[0]->{in_N} = 1;
	my $id = join '-', 'footnote', $_[0]->title, $_[0]->chapter, $_[0]->{footnote}++;
	$_[0]->add_xml_tag( qq|<footnote id="$id"><para>| );
	}

sub end_P   {
	my $entity = $_[0]->get_text_from_current_pad;
	$_[0]->clear_pad;
	$_[0]->{in_P} = 0;

	state $entities = {
		'__ENTITY_TAINT__'    => 'figs/incoming/taint.gif',
		'__ENTITY_TAINT_GREY__'    => 'figs/incoming/taintgrey.gif',
		'__ENTITY_LATEX__'    => 'figs/incoming/entity-latex.png',
		'__ENTITY_XARG__'     => 'figs/incoming/xarg.gif',
		'__ENTITY_XRO__'      => 'figs/incoming/xro.gif',
		'__ENTITY_XU__'       => 'figs/incoming/xu.gif',
		'__ENTITY_XT__'       => 'figs/incoming/xt.gif',
		'__ENTITY_XWIDE__'    => 'figs/incoming/xwide.gif',
		'__ENTITY_DOLLARAT__' => 'figs/incoming/dollarat.gif',
		'__ENTITY_DOLLAR_AT__' => 'figs/incoming/dollarat.gif',
		'__ENTITY_DOLLAR_UNDERSCORE__' => 'figs/incoming/dollarunderscore.gif',
		'__ENTITY_DOLLAR_BANG__' => 'figs/incoming/dollarbang.gif',
		'__ENTITY_DOLLAR_QUESTION__' => 'figs/incoming/dollarquestion.gif',
		};

	unless( exists $entities->{$entity} ) {
		$_[0]->my_warn( "ndefined P<> sequence $entity" );
		$_[0]->add_xml_tag( '<remark>' );
		$_[0]->add_to_pad( "The entity $entity needs to be an inline picture, but we are missing an example" );
		$_[0]->add_xml_tag( '</remark>' );
		$_[0]->add_to_pad( "!!! $entity !!!" );	
		}

	if( $entity eq '__ENTITY_XWIDE__' ) {
		$_[0]->add_xml_tag( '<remark>' );
		$_[0]->add_to_pad( "Our __ENTITY_XWIDE__ image is ugly. Please make a nice looking one :)" );
		$_[0]->add_xml_tag( '</remark>' );
		}
		
	$_[0]->add_xml_tag( '<remark>' );
	$_[0]->add_to_pad( "The entity $entity needs to be an inline picture that looks like $entities->{$entity}" );
	$_[0]->add_xml_tag( '</remark>' );
	
	$_[0]->make_inline_media_object( $entities->{$entity} );

	}
sub start_P {
	$_[0]->emit;
	$_[0]->{in_P} = 1;
	}

sub start_R { $_[0]->add_xml_tag( '<replaceable>' )  }
sub end_R   { $_[0]->add_xml_tag( '</replaceable>' ) }

sub start_S {
	$_[0]->emit;
	$_[0]->{in_S} = 1;
	$_[0]->clear_pad;
	}
sub end_S   {
	my( $self ) = @_;
	my $title = $self->get_text_from_current_pad;
	$self->clear_pad;

	$self->add_xml_tag( '<phrase role="smallcaps">' );

	$title =~ tr[ᴀʙᴄᴅᴇꜰɢʜɪᴊᴋʟᴍɴᴏᴘQʀsᴛᴜᴠᴡxʏᴢ]
	            [abcdefghijklmnopqrstuvwxyz];
	$self->add_to_pad( lc $title );		

	$self->add_xml_tag( '</phrase>' );
	$self->emit;
	$self->clear_pad;
	$self->{in_S} = 0;
	}

sub start_T {
	$_[0]->emit;
	$_[0]->{in_T};
	$_[0]->clear_pad;
	}
sub end_T   {
	state $books = {
		'Learning Perl'     => 'http://my.safaribooksonline.com/book/programming/perl/9781449311063',
		'Intermediate Perl' => 'http://my.safaribooksonline.com/book/programming/perl/0596102062',
		'Mastering Perl'    => 'http://my.safaribooksonline.com/book/programming/perl/9780596527242',
		'Mastering Regular Expressions' => 'http://my.safaribooksonline.com/book/programming/regular-expressions/0596528124',
		'The Perl Cookbook' => 'http://my.safaribooksonline.com/book/programming/perl/0596003137',
		'Star Wars' => 'http://www.starwars.com/',
		'Advanced Programming in the UNIX Environment' => 'http://my.safaribooksonline.com/book/programming/unix/0201433079',
		"Mastering Algorithms with Perl" => 'http://my.safaribooksonline.com/book/programming/perl/1565923987',
		"Modern Perl" => 'http://onyxneon.com/books/modern_perl/',
		"Perl 5 Pocket Reference" => 'http://my.safaribooksonline.com/book/programming/perl/9781449311186',
#		"Object Oriented Perl" => '',
#		"POSIX Programmer's Guide" => '',
#		"Programming perl" => '',
#		"TCP/IP Illustrated" => '',
#		"The Art of Computer Programming" => '',
#		"The Art of Computer Programming, Seminumerical Algorithms" => '',
#		"The Elements of Programming Style" => '',
#		"Algorithms in C" => '',
#		"Introduction to Algorithms" => '',
#		"The Lord of the Rings" => '',
#		"The Unix Programming Environment" => '',
		};

	my( $self ) = @_;
	my $title = $self->get_text_from_current_pad;
	$self->clear_pad;

	$self->add_xml_tag( '<citetitle>' );
	if( exists $books->{$title} ) {
		$self->add_xml_tag( qq(<ulink role="orm:hideurl" url="$books->{$title}">) );
		$self->add_to_pad( $title );
		$self->add_xml_tag( '</ulink>' );
		}
	else {
		$self->add_to_pad( $title );
		}
		

	$self->add_xml_tag( '</citetitle>' );
	$self->emit;
	$self->clear_pad;
	$self->{in_T} = 0;
	}

sub start_U { $_[0]->emit; $_[0]->{in_U} = 1 }
sub end_U   {
	my $text = $_[0]->get_text_from_current_pad;
	$_[0]->clear_pad;

	my( $url, $link_text ) = split /\|/, $text, 2;
	
	if( $url =~ /@/ ) {
		$link_text = $url;
		$url = "mailto:$url";
		}

	if( defined $link_text ) {
		$_[0]->add_xml_tag( qq|<ulink url="$url">$link_text</ulink>| );		
		}
	else {
		$_[0]->add_xml_tag( qq|<ulink url="$url" />| );
		}
	$_[0]->emit;
	$_[0]->{in_U} = 0;
	}

sub start_V {
	$_[0]->emit;
	$_[0]->{in_V} = 1;
	$_[0]->clear_pad;
	}
sub end_V   {
	my( $self ) = @_;
	my $variable = $self->get_text_from_current_pad;
	$_[0]{'in_V'} = 0;
	
	$self->add_xml_tag( '<literal moreinfo="none">' );
	$self->add_to_pad( $variable );
	$self->add_xml_tag( '</literal>' );


	$_[0]->add_xml_tag( '<indexterm type="variable">' );

	$_[0]->add_xml_tag( '<primary>' );
	$_[0]->add_to_pad( $variable );
	$_[0]->add_xml_tag( '</primary>' );
	$_[0]->add_xml_tag( '<secondary>' );
	$_[0]->add_to_pad( $variable );
	$_[0]->add_xml_tag( '</secondary>' );

	$_[0]->add_xml_tag( '</indexterm>' );
	}

# http://docbook.org/tdg/en/html/indexterm.html
# <!DOCTYPE para PUBLIC "-//OASIS//DTD DocBook XML V4.1.2//EN"
#           "http://www.oasis-open.org/docbook/xml/4.1.2/docbookx.dtd">
# <para>
# The Tiger<indexterm>
# <primary>Big Cats</primary>
# <secondary>Tigers</secondary></indexterm>
# is a very large cat indeed.
# </para>
sub start_X { 
	$_[0]->emit;
	$_[0]->{in_X} = 1;	
	}
sub end_X   {
	my $text = $_[0]->get_text_from_current_pad;
	$text =~ s/\A\s+//;
	$text =~ s/\s+\z//;
	$text =~ s/\s+/ /g;

	$_[0]->clear_pad;
	my( $terms, $see_also ) = split /\s*;;\s*/, $text;

	my @terms = split /\s*;\s*/, $terms; local $" = '><';

	return unless @terms;
	$_[0]->my_warn( "More than three levels in index term [$text]!" ) 
		if 0; #@terms > 3; XXX turn on later
		
	$_[0]->add_xml_tag( '<indexterm>' );

	$_[0]->add_xml_tag( '<primary>' );
	$_[0]->add_to_pad( $terms[0] );
	$_[0]->escape_and_emit;
	$_[0]->add_xml_tag( '</primary>' );

	if( @terms > 1 ) {
		$_[0]->add_xml_tag( '<secondary>' );
		$_[0]->add_to_pad( $terms[1] );
		$_[0]->escape_and_emit;
		$_[0]->add_xml_tag( '</secondary>' );
		}		

	if( @terms > 2 ) {
		$_[0]->add_xml_tag( '<tertiary>' );
		$_[0]->add_to_pad( $terms[2] );
		$_[0]->escape_and_emit;
		$_[0]->add_xml_tag( '</tertiary>' );
		}		

	if( defined $see_also ) {
		$_[0]->add_xml_tag( '<seealso>' );
		$_[0]->escape_and_emit( $see_also );
		$_[0]->add_xml_tag( '</seealso>' );
		}		
	
	$_[0]->add_xml_tag( '</indexterm>' );
	$_[0]->{in_X} = 0;
	}

sub start_Y { $_[0]->{in_Y} = 1; $_[0]->add_xml_tag( '<superscript>' );  }
sub end_Y   { $_[0]->add_xml_tag( '</superscript>' ); $_[0]->{in_Y} = 0; }

sub start_Z { $_[0]->escape_and_emit; $_[0]->{in_Z} = 1;  }
sub end_Z   { $_[0]->clear_pad; $_[0]->{in_Z} = 0; }

sub my_warn {
	my( $self, @texts ) = @_;

	local $" = '';
	warn "$self->{source_filename}:$self->{line_count} @texts\n";
	}

sub handle_text {
	my( $self, $text ) = @_;

	$self->escape_text( \$text );
	if( $self->{in_verbatim} && $text !~ /^\s+B&lt;-0xR/m ) {
		$text =~ s"B &lt;&lt;(?!&lt;) (.*?) &gt&gt; "<userinput>$1</userinput>"xg; # hack for now
		$text =~ s|R &lt; (.*?) &gt; |<replaceable>$1</replaceable>|xg;
		$text =~ s"B &lt; (.*?) &gt; "<userinput>$1</userinput>"xg; # hack for now
		$text =~ s"BZ&lt;&gt;"B"xg; # hack for now
		}
	if( $self->{in_figure} ) {
		$self->add_to_pad( ' ' ); # separate filenames
		}
	$self->add_to_pad( $text );

	unless( $self->dont_escape ) {
		$self->make_curly_quotes;
		$self->make_em_dashes;
		$self->make_ellipses;
		}
	}

sub dont_escape {
	my $self = shift;
	$self->{in_verbatim} || $self->{in_C} || $self->{in_X}
	}

sub escape_text {
	my( $self, $text_ref ) = @_;

	$$text_ref =~ s/&(?!(?:lt|gt|amp|#x(?:[a-f0-9]+));)/&amp;/gi;
	$$text_ref =~ s/</&lt;/g;
	$$text_ref =~ s/>/&gt;/g;

	return 1;
	}

sub make_curly_quotes {
	my( $self ) = @_;

	my $pad  = $self->get_pad;
	my $text = $self->{$pad};

	require Tie::Cycle;

	tie my $cycle, 'Tie::Cycle', [ qw( &#x201C; &#x201D; ) ];

	1 while $text =~ s/"/$cycle/;

	# single quotes are weird because they mostly don't
	# come in pairs
	$text =~ s/(^|\s)'/$1&#x2018;/g;
	$text =~ s/'/&#x2019;/g;

	$self->{$pad} = $text;

	return 1;
	}

sub make_em_dashes {
	my( $self ) = @_;
	return if $self->{in_verbose};
	my $pad  = $self->get_pad;
	$_[0]->{$pad} =~ s/--/&#x2014;/g;
	return 1;
	}

sub make_ellipses {
	my( $self ) = @_;
	my $pad  = $self->get_pad;
	$self->{$pad} =~ s/\Q.../&#x2026;/g;
	return 1;
	}

BEGIN {
require Pod::Simple::BlackBox;

package Pod::Simple::BlackBox;

sub InFancyDash {
	return <<"CODE_NUMBERS";
2010 2015
2212
FE58
CODE_NUMBERS
	}

sub _ponder_Verbatim {
	my ($self,$para) = @_;
	DEBUG and print STDERR " giving verbatim treatment...\n";

	$para->[1]{'xml:space'} = 'preserve';

	# strip trailing whitespace	
	s/\h+$// foreach ( @$para[ 2 .. $#$para ] );
	
	# remove leading blank lines	
	while( $para->[2] =~ m/ \A \h* $ /x ) {
		splice @$para, 2, 1, ();
		}

	# find the leading whitespace on the first line
	my( $leader ) = $para->[2] =~ m/ \A (\h+) /x;
	
	foreach my $line ( @$para[ 2 .. $#$para ] ) {
		$line =~ s/ \A \Q$leader\E //x;

		$self->my_warn(
				"tab in code listing!\n\t$line",
			) if $line =~ /\t/;
			
		if( $line =~ /\p{InFancyDash}/ ) {
			# only if it's not in a comment
			my $post_comment = $line =~ s/.*?#//r;
			
			$self->my_warn(
					"fancy dash in code listing!\n\t$line",
				) unless $post_comment =~ /\p{InFancyDash}/; # −
			}
  		}

  # Now the VerbatimFormatted hoodoo...
  if( $self->{'accept_codes'} and
      $self->{'accept_codes'}{'VerbatimFormatted'}
  ) {
    while(@$para > 3 and $para->[-1] !~ m/\S/) { pop @$para }
     # Kill any number of terminal newlines
    $self->_verbatim_format($para);
  } elsif ($self->{'codes_in_verbatim'}) {
    push @$para,
    @{$self->_make_treelet(
      join("\n", splice(@$para, 2)),
      $para->[1]{'start_line'}, $para->[1]{'xml:space'}
    )};
    $para->[-1] =~ s/\n+$//s; # Kill any number of terminal newlines
  } else {
    push @$para, join "\n", splice(@$para, 2) if @$para > 3;
    $para->[-1] =~ s/\n+$//s; # Kill any number of terminal newlines
  }
  return;
}

}

BEGIN {
use Pod::PseudoPod;

package Pod::PseudoPod;

sub _ponder_begin {
  my ($self,$para,$curr_open,$paras) = @_;

  unless ($para->[2] =~ /^\s*(?:table|sidebar|figure|listing|production)/) {
    return $self->SUPER::_ponder_begin($para,$curr_open,$paras);
  }

  my $content = join ' ', splice @$para, 2;
  $content =~ s/^\s+//s;
  $content =~ s/\s+$//s;

  my ($target, $title) = $content =~ m/^(\S+)\s*(.*)$/;
  $title =~ s/^(picture|html)\s*// if ($target eq 'table');
  $para->[1]{'title'} = $title if ($title);
  $para->[1]{'target'} = $target;  # without any ':'

  return 1 unless $self->{'accept_targets'}{$target};

  $para->[0] = '=for';  # Just what we happen to call these, internally
  $para->[1]{'~really'} ||= '=begin';
#  $para->[1]{'~ignore'}  = 0;
  $para->[1]{'~resolve'} = 1;

  push @$curr_open, $para;
  $self->{'content_seen'} ||= 1;
  $self->_handle_element_start($target, $para->[1]);

  return 1;
}

sub _ponder_end {
  my ($self,$para,$curr_open,$paras) = @_;
  my $content = join ' ', splice @$para, 2;
  $content =~ s/^\s+//s;
  $content =~ s/\s+$//s;
  DEBUG and print "Ogling '=end $content' directive\n";
  
  unless(length($content)) {
    if (@$curr_open and $curr_open->[-1][1]{'~really'} eq '=for') {
      # =for allows an empty =end directive
      $content = $curr_open->[-1][1]{'target'};
    } else {
      # Everything else should complain about an empty =end directive
      my $complaint = "'=end' without a target?";
      if ( @$curr_open and $curr_open->[-1][0] eq '=for' ) {
        $complaint .= " (Should be \"=end " . $curr_open->[-1][1]{'target'} . '")';
      }
      $self->whine( $para->[1]{'start_line'}, $complaint);
      DEBUG and print "Ignoring targetless =end\n";
      return 1;
    }
  }
  
  unless($content =~ m/^\S+$/) {  # i.e., unless it's one word
    $self->whine(
      $para->[1]{'start_line'},
      "'=end $content' is invalid.  (Stack: "
      . $self->_dump_curr_open() . ')'
    );
    DEBUG and print "Ignoring mistargetted =end $content\n";
    return 1;
  }
  
  $self->_ponder_row_end($para,$curr_open,$paras) if $content eq 'table';

  unless(@$curr_open and $curr_open->[-1][0] eq '=for') {
    $self->whine(
      $para->[1]{'start_line'},
      "=end $content without matching =begin.  (Stack: "
      . $self->_dump_curr_open() . ')'
    );
    DEBUG and print "Ignoring mistargetted =end $content\n";
    return 1;
  }
  
  unless($content eq $curr_open->[-1][1]{'target'}) {
    if ($content eq 'for' and $curr_open->[-1][1]{'~really'} eq '=for') {
      # =for allows a "=end for" directive
      $content = $curr_open->[-1][1]{'target'};
    } else {
      $self->whine(
        $para->[1]{'start_line'},
        "=end $content doesn't match =begin " 
        . $curr_open->[-1][1]{'target'}
        . ".  (Stack: "
        . $self->_dump_curr_open() . ')'
      );
      DEBUG and print "Ignoring mistargetted =end $content at line $para->[1]{'start_line'}\n";
      return 1;
    }
  }

  # Else it's okay to close...
  if(grep $_->[1]{'~ignore'}, @$curr_open) {
    DEBUG > 1 and print "Not firing any event for this =end $content because in an ignored region\n";
    # And that may be because of this to-be-closed =for region, or some
    #  other one, but it doesn't matter.
  } else {
    $curr_open->[-1][1]{'start_line'} = $para->[1]{'start_line'};
      # what's that for?
    
    $self->{'content_seen'} ||= 1;
    if ($content eq 'production' or $content eq 'table' or $content eq 'sidebar' or $content eq 'figure' or $content eq 'listing') {
      $self->_handle_element_end( $content );
    } else {
      $self->_handle_element_end( 'for', { 'target' => $content } );
    }
  }
  DEBUG > 1 and print "Popping $curr_open->[-1][0] $curr_open->[-1][1]{'target'} because of =end $content\n";
  pop @$curr_open;

  return 1;
} 
}



BEGIN {

# override _treat_Es so I can localize e2char
sub _treat_Es {
	my $self = shift;

	require Pod::Escapes;
	local *Pod::Escapes::e2char = *e2char_tagged_text;

	$self->SUPER::_treat_Es( @_ );
	}

sub e2char_tagged_text {
	package Pod::Escapes;

	my $in = shift;
	return unless defined $in and length $in;

	   if( $in =~ m/^(0[0-7]*)$/ )         { $in = oct $in; }
	elsif( $in =~ m/^0?x([0-9a-fA-F]+)$/ ) { $in = hex $1;  }

	if( $NOT_ASCII ) {
		unless( $in =~ m/^\d+$/ ) {
			$in = $Name2character{$in};
			return unless defined $in;
			$in = ord $in;
	    	}

		return $Code2USASCII{$in}
			|| $Latin1Code_to_fallback{$in}
			|| $FAR_CHAR;
		}

 	if( defined $Name2character_number{$in} and $Name2character_number{$in} < 127 ) {
 		return "&$in;";
 		}
	elsif( defined $Name2character_number{$in} ) {
		# this needs to be fixed width because I want to look for
		# it in a negative lookbehind
		return sprintf '&#x%04x;', $Name2character_number{$in};
		}
	else {
		return '???';
		}

	}
}

=head1 TO DO


=head1 SEE ALSO

L<Pod::PseudoPod>, L<Pod::Simple>

=head1 SOURCE AVAILABILITY

This source is in Github:

	http://github.com/briandfoy/Pod-DocBook

If, for some reason, I disappear from the world, one of the other
members of the project can shepherd this module appropriately.

=head1 AUTHOR

brian d foy, C<< <bdfoy@cpan.org> >>

=head1 COPYRIGHT AND LICENSE

Copyright (c) 2010, brian d foy, All Rights Reserved.

You may redistribute this under the same terms as Perl itself.

=cut

sub _ponder_paragraph_buffer {

  # Para-token types as found in the buffer.
  #   ~Verbatim, ~Para, ~end, =head1..4, =for, =begin, =end,
  #   =over, =back, =item
  #   and the null =pod (to be complained about if over one line)
  #
  # "~data" paragraphs are something we generate at this level, depending on
  # a currently open =over region

  # Events fired:  Begin and end for:
  #                   directivename (like head1 .. head4), item, extend,
  #                   for (from =begin...=end, =for),
  #                   over-bullet, over-number, over-text, over-block,
  #                   item-bullet, item-number, item-text,
  #                   Document,
  #                   Data, Para, Verbatim
  #                   B, C, longdirname (TODO -- wha?), etc. for all directives
  #

  my $self = $_[0];
  my $paras;
  return unless @{$paras = $self->{'paras'}};
  my $curr_open = ($self->{'curr_open'} ||= []);

  DEBUG > 10 and print "# Paragraph buffer: <<", pretty($paras), ">>\n";

  # We have something in our buffer.  So apparently the document has started.
  unless($self->{'doc_has_started'}) {
    $self->{'doc_has_started'} = 1;

    my $starting_contentless;
    $starting_contentless =
     (
       !@$curr_open
       and @$paras and ! grep $_->[0] ne '~end', @$paras
        # i.e., if the paras is all ~ends
     )
    ;
    DEBUG and print "# Starting ",
      $starting_contentless ? 'contentless' : 'contentful',
      " document\n"
    ;

    $self->_handle_element_start('Document',
      {
        'start_line' => $paras->[0][1]{'start_line'},
        $starting_contentless ? ( 'contentless' => 1 ) : (),
      },
    );
  }

  my($para, $para_type);
  while(@$paras) {
    last if @$paras == 1 and
      ( $paras->[0][0] eq '=over' or $paras->[0][0] eq '~Verbatim'
        or $paras->[0][0] eq '=item' )
    ;
    # Those're the three kinds of paragraphs that require lookahead.
    #   Actually, an "=item Foo" inside an <over type=text> region
    #   and any =item inside an <over type=block> region (rare)
    #   don't require any lookahead, but all others (bullets
    #   and numbers) do.

# TODO: winge about many kinds of directives in non-resolving =for regions?
# TODO: many?  like what?  =head1 etc?

    $para = shift @$paras;
    $para_type = $para->[0];

    DEBUG > 1 and print "Pondering a $para_type paragraph, given the stack: (",
      $self->_dump_curr_open(), ")\n";

    if($para_type eq '=for') {
      next if $self->_ponder_for($para,$curr_open,$paras);
    } elsif($para_type eq '=begin') {
      next if $self->_ponder_begin($para,$curr_open,$paras);
    } elsif($para_type eq '=end') {
      next if $self->_ponder_end($para,$curr_open,$paras);
    } elsif($para_type eq '~end') { # The virtual end-document signal
      next if $self->_ponder_doc_end($para,$curr_open,$paras);
    }


    # ~ ~ ~ ~ ~ ~ ~ ~ ~ ~ ~ ~ ~ ~ ~ ~ ~ ~ ~ ~ ~ ~ ~ ~ ~ ~ ~ ~ ~ ~ ~ ~ ~ ~ ~ ~
    #~ ~ ~ ~ ~ ~ ~ ~ ~ ~ ~ ~ ~ ~ ~ ~ ~ ~ ~ ~ ~ ~ ~ ~ ~ ~ ~ ~ ~ ~ ~ ~ ~ ~ ~ ~ ~
    if(grep $_->[1]{'~ignore'}, @$curr_open) {
      DEBUG > 1 and
       print "Skipping $para_type paragraph because in ignore mode.\n";
      next;
    }
    #~ ~ ~ ~ ~ ~ ~ ~ ~ ~ ~ ~ ~ ~ ~ ~ ~ ~ ~ ~ ~ ~ ~ ~ ~ ~ ~ ~ ~ ~ ~ ~ ~ ~ ~ ~ ~
    # ~ ~ ~ ~ ~ ~ ~ ~ ~ ~ ~ ~ ~ ~ ~ ~ ~ ~ ~ ~ ~ ~ ~ ~ ~ ~ ~ ~ ~ ~ ~ ~ ~ ~ ~ ~

    if($para_type eq '=pod') {
      $self->_ponder_pod($para,$curr_open,$paras);
    } elsif($para_type eq '=over') {
      next if $self->_ponder_over($para,$curr_open,$paras);
    } elsif($para_type eq '=back') {
      next if $self->_ponder_back($para,$curr_open,$paras);
    } elsif($para_type eq '=row') {
      next if $self->_ponder_row_start($para,$curr_open,$paras);

    } elsif( $para_type eq '=headrow'){
    	$self->start_headrow;
    } elsif( $para_type eq '=bodyrows') {
    	$self->start_bodyrows;
    	}

    else {
      # All non-magical codes!!!

      # Here we start using $para_type for our own twisted purposes, to
      #  mean how it should get treated, not as what the element name
      #  should be.

      DEBUG > 1 and print "Pondering non-magical $para_type\n";

      # In tables, the start of a headrow or bodyrow also terminates an
      # existing open row.
      if($para_type eq '=headrow' || $para_type eq '=bodyrows') {
        $self->_ponder_row_end($para,$curr_open,$paras);
      }

      # Enforce some =headN discipline
      if($para_type =~ m/^=head\d$/s
         and ! $self->{'accept_heads_anywhere'}
         and @$curr_open
         and $curr_open->[-1][0] eq '=over'
      ) {
        DEBUG > 2 and print "'=$para_type' inside an '=over'!\n";
        $self->whine(
          $para->[1]{'start_line'},
          "You forgot a '=back' before '$para_type'"
        );
        unshift @$paras, ['=back', {}, ''], $para;   # close the =over
        next;
      }


      if($para_type eq '=item') {
        next if $self->_ponder_item($para,$curr_open,$paras);
        $para_type = 'Plain';
        # Now fall thru and process it.

      } elsif($para_type eq '=extend') {
        # Well, might as well implement it here.
        $self->_ponder_extend($para);
        next;  # and skip
      } elsif($para_type eq '=encoding') {
        # Not actually acted on here, but we catch errors here.
        $self->_handle_encoding_second_level($para);

        next;  # and skip
      } elsif($para_type eq '~Verbatim') {
        $para->[0] = 'Verbatim';
        $para_type = '?Verbatim';
      } elsif($para_type eq '~Para') {
        $para->[0] = 'Para';
        $para_type = '?Plain';
      } elsif($para_type eq 'Data') {
        $para->[0] = 'Data';
        $para_type = '?Data';
      } elsif( $para_type =~ s/^=//s
        and defined( $para_type = $self->{'accept_directives'}{$para_type} )
      ) {
        DEBUG > 1 and print " Pondering known directive ${$para}[0] as $para_type\n";
      } else {
        # An unknown directive!
        DEBUG > 1 and printf "Unhandled directive %s (Handled: %s)\n",
         $para->[0], join(' ', sort keys %{$self->{'accept_directives'}} )
        ;
        $self->whine(
          $para->[1]{'start_line'},
          "Unknown directive: $para->[0]"
        );

        # And maybe treat it as text instead of just letting it go?
        next;
      }

      if($para_type =~ s/^\?//s) {
        if(! @$curr_open) {  # usual case
          DEBUG and print "Treating $para_type paragraph as such because stack is empty.\n";
        } else {
          my @fors = grep $_->[0] eq '=for', @$curr_open;
          DEBUG > 1 and print "Containing fors: ",
            join(',', map $_->[1]{'target'}, @fors), "\n";

          if(! @fors) {
            DEBUG and print "Treating $para_type paragraph as such because stack has no =for's\n";

          #} elsif(grep $_->[1]{'~resolve'}, @fors) {
          #} elsif(not grep !$_->[1]{'~resolve'}, @fors) {
          } elsif( $fors[-1][1]{'~resolve'} ) {
            # Look to the immediately containing for

            if($para_type eq 'Data') {
              DEBUG and print "Treating Data paragraph as Plain/Verbatim because the containing =for ($fors[-1][1]{'target'}) is a resolver\n";
              $para->[0] = 'Para';
              $para_type = 'Plain';
            } else {
              DEBUG and print "Treating $para_type paragraph as such because the containing =for ($fors[-1][1]{'target'}) is a resolver\n";
            }
          } else {
            DEBUG and print "Treating $para_type paragraph as Data because the containing =for ($fors[-1][1]{'target'}) is a non-resolver\n";
            $para->[0] = $para_type = 'Data';
          }
        }
      }

      #~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
      if($para_type eq 'Plain') {
        $self->_ponder_Plain($para);
      } elsif($para_type eq 'Verbatim') {
        $self->_ponder_Verbatim($para);
      } elsif($para_type eq 'Data') {
        $self->_ponder_Data($para);
      } else {
        die "\$para type is $para_type -- how did that happen?";
        # Shouldn't happen.
      }

      #~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
      $para->[0] =~ s/^[~=]//s;

      DEBUG and print "\n", Pod::Simple::BlackBox::pretty($para), "\n";

      # traverse the treelet (which might well be just one string scalar)
      $self->{'content_seen'} ||= 1;
      $self->_traverse_treelet_bit(@$para);
    }
  }

  return;
}

1;
